# -*- coding: latin-1 -*-
'''
outputManager.py

Copyright 2007 Mariano Nuez Di Croce @ CYBSEC

This file is part of sapyto, http://www.cybsec.com/EN/research/tools/sapyto.php

sapyto is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

sapyto is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sapyto; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''
import os,sys
from core.misc import factory

class outputManager:
	'''
	This class manages output. 
	It has a list of output plugins and sends the events to every plugin on that list.
	
	@author: Andres Riancho ( andres.riancho@gmail.com )
	'''
	
	def __init__(self):
		self._outputPluginList = []
		self._pluginsOptions = {}
		self._levels = ['info',]
	
	def setLevels(self, levels):
		'''
			Sets levels of output verbosity.
			
			@parameter levels: List containing levels of output verbosity.
				'info' = information messages
				'debug' = debug messages
				'error' = error messages
				'vuln' = vulnerability
		'''
		self._levels = levels

	def getLevels(self):
		'''
			Gets levels of output verbosity.
		'''
		return self._levels

	def quiet(self):
		''' 
		Stop outputing data.
		'''
		self._tmp = self._levels
		self._levels = []
		
	def talk(self):
		''' 
		Restarts outputing data.
		'''
		self._levels = self._tmp
	
	
	def _addOutputPlugin(self, OutputPluginName ):
		'''
		Takes a string with the OutputPluginName, creates the object and adds it to the OutputPluginName
		
		@parameter OutputPluginName: The name of the plugin to add to the list.
		@return: No value is returned.
		'''
		if OutputPluginName == 'all':
			fileList = [ f for f in os.listdir('plugins' +os.path.sep +'output'+os.path.sep) ]    
			strReqPlugins = [ os.path.splitext(f)[0] for f in fileList if os.path.splitext(f)[1] == '.py']
			strReqPlugins.remove ( '__init__' )
			
			for pluginName in strReqPlugins:
				plugin = factory( 'plugins.output.' + pluginName )
				
				if pluginName in self._pluginsOptions.keys():
					plugin.setOptions( self._pluginsOptions[pluginName] )
				
				# Append the plugin to the list
				self._outputPluginList.append( plugin )
		
		else:
				plugin = factory( 'plugins.output.' + OutputPluginName )
				if OutputPluginName in self._pluginsOptions.keys():
					plugin.setOptions( self._pluginsOptions[OutputPluginName] )

					# Append the plugin to the list
				self._outputPluginList.append( plugin )
				
	def debug(self, message, newLine = True ):
		'''
		Sends a debug message to every output plugin on the list.
		
		@parameter message: Message that is sent.
		'''
		if 'debug' in self._levels:
			for oPlugin in self._outputPluginList:
				oPlugin.debug(message, newLine )
	
	def information(self, message, newLine = True ):
		'''
		Sends a informational message to every output plugin on the list.
		
		@parameter message: Message that is sent.
		'''
		if 'info' in self._levels:
			for oPlugin in self._outputPluginList:
				oPlugin.information( message, newLine )
			
	def error(self, message, newLine = True ):
		'''
		Sends an error message to every output plugin on the list.
		
		@parameter message: Message that is sent.
		'''
		if 'error' in self._levels:
			for oPlugin in self._outputPluginList:
				oPlugin.error(message, newLine )

			
	def vulnerability(self, message, newLine = True ):
		'''
		Sends a vulnerability message to every output plugin on the list.
		
		@parameter message: Message that is sent.
		'''
		if 'vuln' in self._levels:
			for oPlugin in self._outputPluginList:
				oPlugin.vulnerability( message, newLine )

	def console( self, message, newLine = True ):
		'''
		This method is used by the sapyto console to print messages to the outside.
		'''
		for oPlugin in self._outputPluginList:
			oPlugin.console( message, newLine )
			
	def setOutputPlugins( self, OutputPlugins ):
		'''
		@parameter OutputPlugins: A list with the names of Output Plugins that will be used.
		@return: No value is returned.
		''' 	
		self._outputPluginList = []
		self._outputPlugins = OutputPlugins
		
		for pluginName in self._outputPlugins:
			out._addOutputPlugin( pluginName )	
		
		out.debug('Exiting _createOM()' )
		
	def setPluginOptions(self, pluginName, PluginsOptions ):
		'''
		@parameter PluginsOptions: A tuple with a string and a dictionary with the options for a plugin. For example:\
		{ console:{'verbosity':7} }
			
		@return: No value is returned.
		''' 		
		self._pluginsOptions[pluginName] = PluginsOptions

out = outputManager()
