# -*- coding: latin-1 -*-
'''
sapytoCore.py

Copyright 2007 Mariano Nuez Di Croce @ CYBSEC

This file is part of sapyto, http://www.cybsec.com/EN/research/tools/sapyto.php

sapyto is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

sapyto is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sapyto; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
'''
import os,sys
import core.output.outputManager as om
from core.sapytoException import sapytoException
from core.misc import *
import traceback
import core.kb.knowledgeBase as kb
import plugins.audit.ping as ping
import saprfc
import copy
import re

class sapytoCore:
	'''
	This is the core of the framework, it calls all plugins, handles exceptions,
	coordinates all the work, etc.
	 
	@author: Mariano Nuez Di Croce <mnunez@cybsec.com>
	'''

	def __init__(self):
	
		
		# A map with plugin types as keys and a list of plugin names as values
		self._strPlugins = {'audit':[],'attack':[],'output':[]}
		
		# A map with plugin types as keys and a list of plugin instances as values
		self._plugins = {'audit':[],'attack':[],'output':[]}

		self._pluginsOptions = {}
		
		self._autoDependencies = True
		self._conn = None
		self._targets = {}
		
		'''
		# Debugging
		self._targets['type'] = 'SAP'
		self._targets['ashost'] = '172.1.1.2'
		self._targets['user'] = 'sap*'
		self._targets['passwd'] = 'minisap'
		self._targets['client'] = '000'
		
		self._targets['type'] = 'EXT'
		self._targets['gwhost'] = '172.1.1.2'
		self._targets['gwserv'] = 'sapgw00'
		self._targets['tpname'] = 'cyb_ext1'
		'''
	
	def getConn(self):
		'''
		Returns connection object
		'''
		return self._conn
		
	
	def getTarget(self):
		'''
		Retuns sapyto targets.
		'''
		return self._targets
	
	def setTarget(self, target):
		'''
		Add a target to sapyto targets.
		'''
		tmp = {}
		tmp['ashost'] = target['ashost']['default']
		tmp['client'] = target['client']['default']
		tmp['sysnr'] = target['sysnr']['default']
		tmp['gwhost'] = target['gwhost']['default']
		tmp['gwserv'] = target['gwserv']['default']
		tmp['user'] = target['user']['default']
		tmp['passwd'] = target['passwd']['default']
		tmp['tpname'] = target['tpname']['default']
		tmp['type'] = target['type']['default']
		self._targets = tmp
		self._conn = self._buildConnection(self._targets)
	
	def start(self):
		'''
		Starts the work.
		
		@return: No value is returned.
		''' 
		try:
				
			if not self._targets:
				om.out.information('No target defined.')
				return
			
			self._initPlugins()
			
			# Check connection with target, before starting plugins.
			p = ping.ping()
			p.setConn(self._conn)
			#quietly			
			om.out.quiet()
			res = p.run()
			om.out.talk()
			if res:
				self._audit()
				self._attack()
			else:
				om.out.information('No connection with target. Exiting...')
			
		except sapytoException, e:
			#om.out.error( str(e) )
			#om.out.information('An unhandled exception occurred')
			raise e
		except KeyboardInterrupt, e:
			# I wont handle this. 
			# The user interface must know what to do with it
			raise e

	def _buildConnection(self, targets):
		'''
		This method creates the connection to target.
		
		@parameter targets: Target information
		@returns: Connection object
		'''
		if targets['type'] == 'SAP':
			conn = saprfc.conn(ashost = targets['ashost'], client = targets['client'], user = targets['user'], passwd = targets['passwd'], type= '3', sysnr=targets['sysnr'])
		else:
			conn = saprfc.conn(gwhost = targets['gwhost'], gwserv = targets['gwserv'], tpname = targets['tpname'], type= 'E')
		
		return conn
		
	
	def _rPlugFactory( self, strReqPlugins, PluginType ):
		'''
		This method creates the requested modules list.
		
		@parameter strReqPlugins: A string list with the requested plugins to be executed.
		@parameter PluginType: [audit|discovery|grep]
		@return: A list with plugins to be executed, this list is ordered using the exec priority.
		'''		
		requestedPluginsList = []
		
		if 'all' in strReqPlugins:
			fileList = [ f for f in os.listdir('plugins' + os.path.sep+ PluginType + os.path.sep ) ]    
			allPlugins = [ os.path.splitext(f)[0] for f in fileList if os.path.splitext(f)[1] == '.py' ]
			allPlugins.remove ( '__init__' )
			
			if len ( strReqPlugins ) != 1:
				unwantedPlugins = [ x[1:] for x in strReqPlugins if x[0] =='!' ]
				strReqPlugins = list( set(allPlugins) - set(unwantedPlugins) ) #bleh! v2
			else:
				strReqPlugins = allPlugins
			
			# Update the plugin list
			# This update is usefull for cases where the user selected "all" plugins,
			# the self._strPlugins[PluginType] is useless if it says 'all'.
			self._strPlugins[PluginType] = strReqPlugins
				
		for pluginName in strReqPlugins:
			plugin = factory( 'plugins.' + PluginType + '.' + pluginName )

			# Now we set the plugin options
			if pluginName in self._pluginsOptions.keys():
				plugin.setOptions( self._pluginsOptions[pluginName] )
				
			# Append the plugin to the list
			requestedPluginsList.append ( plugin )
			
		return requestedPluginsList
	
	def _initPlugins( self ):
		# First, create an instance of each requested plugin and add it to the plugin list
		# Plugins are added taking care of plugin dependencies
		self._plugins['audit'] = self._rPlugFactory( self._strPlugins['audit'] , 'audit')
		self._plugins['attack'] = self._rPlugFactory( self._strPlugins['attack'] , 'attack')		
			
	def _audit(self):
		om.out.debug('Called _audit()' )
		for plugin in self._plugins['audit']:
			om.out.information('')
			plugin.setConn(self._conn)
			try:
				plugin.run()
			except sapytoException:
				continue
	
	def _attack(self):
		om.out.debug('Called _attack()' )
		for plugin in self._plugins['attack']:
			#om.out.information('Starting ' + plugin.getName() + ' plugin execution.')
			om.out.information('')
			plugin.setConn(self._conn)
			try:
				plugin.run()
			except sapytoException:
				continue
				
	def setPluginOptions(self, pluginName, PluginsOptions ):
		'''
		@parameter PluginsOptions: A tuple with a string and a dictionary with the options for a plugin. For example:\
		{ googlespider_plugin:{'LICENSE_KEY':'AAAA'} }
			
		@return: No value is returned.
		''' 		
		self._pluginsOptions[pluginName] = PluginsOptions
	
	def setPlugins( self, pluginNames, pluginType ):
		'''
		This method sets the plugins that sapytoCore is going to use. 
		
		@parameter pluginNames: A list with the names of the Plugins that will be runned.
		@parameter pluginType: The type of the plugin.
		@return: None
		'''
		setMap = {'audit':self._setAuditPlugins, 'output':self._setOutputPlugins, 'attack': self._setAttackPlugins}
		
		func = setMap[ pluginType ]
		func( pluginNames )
		
	def getPluginTypes( self ):
		'''
		@return: A list with all plugin types.
		'''
		pluginTypes = [ f for f in os.listdir('plugins/') if f.count('.py') == 0 ]
		#pluginTypes.remove( 'attack' )
		return pluginTypes
		
	def _setOutputPlugins( self, outputPlugins ):
		'''
		@parameter outputPlugins: A list with the names of output Plugins that will be runned.
		@return: No value is returned.
		'''
		self._strPlugins['output'] = outputPlugins
		
	def _setAuditPlugins( self, AuditPlugins ):
		'''
		@parameter AuditPlugins: A list with the names of Audit Plugins that will be runned.
		@return: No value is returned.
		''' 		
		self._strPlugins['audit'] = AuditPlugins
		
	def _setAttackPlugins( self, AttackPlugins):
		'''
		@parameter AttackPlugins: A list with the names of Attack Plugins that will be used.
		@return: No value is returned.
		''' 	
		self._strPlugins['attack'] = AttackPlugins
		
	def getEnabledPlugins(self, pluginType):
		'''
		@parameter pluginType: Type of plugin
		Gets already enabled plugins of pluginType.
		'''
		return self._strPlugins[pluginType]
	
	def getPluginList( self, PluginType ):
		'''
		@return: A string list of the names of all available exploit plugins.
		'''
		fileList = [ f for f in os.listdir('plugins/' + PluginType +'/') ]    
		strReqPlugins = [ os.path.splitext(f)[0] for f in fileList if os.path.splitext(f)[1] == '.py' ]
		strReqPlugins.remove ( '__init__' )
		return strReqPlugins
		
	def getPluginInstance( self, pluginName ):
		'''
		@return: An instance of a plugin.
		'''
		# First we need to find the plugin.
		pluginTypes = [ f for f in os.listdir('plugins/') ]
		pluginTypes = [ f for f in pluginTypes if os.path.isdir( os.path.join('plugins/',f) ) ]

		for pluginType in pluginTypes:
			fileList = [ f for f in os.listdir('plugins/' + pluginType +'/') ]    
			fileList = [ os.path.splitext(f)[0] for f in fileList if os.path.splitext(f)[1] == '.py' ]
			fileList.remove ( '__init__' )
			if pluginName in fileList:
				ModuleName = 'plugins.'+pluginType+ '.'+ pluginName
				__import__(ModuleName)
				aModule = sys.modules[ModuleName]
				className = ModuleName.split('.')[len(ModuleName.split('.'))-1]
				aClass = getattr( aModule , className )
				plugin = apply(aClass, ())
				if pluginName in self._pluginsOptions.keys():
					plugin.setOptions( self._pluginsOptions[pluginName] )
				return plugin
				
		raise sapytoException('Plugin not found')
	
	def setAutoDependencies( self, autodep ):
		'''
		Set auto dependencies resolution.
		
		@parameter autodep: Boolean
		'''
		self._autoDependencies = autodep
	
	def getAutoDependencies( self ): return self._autoDependencies

	def _getPluginsRecursive(self, dir):
		res = []
		fileList = [f for f in os.listdir(dir)]
		for d in fileList:
			if os.path.isdir(dir + os.path.sep + d):
				res.extend(self._getPluginsRecursive(dir + os.path.sep + d))
			else:
				if d != '__init__.py':
					res.append(dir + '.' + d)		
		res = [f.replace(os.path.sep, '.').strip() for f in res]
		return res

	def validateTargets(self, targets):
		'''
		Validates plugins targets with information obtained from RFC_SYSTEM_INFO on remote RFC server.
		Regular Expressions used.
		
		@parameter targets: Targets to validate
		@returns: Boolean for defining if target can be audited by the plugin
		'''
		# No checking for non-checking plugins
		if targets == {}:
			return True
		
		# Verify if sapinfo was called before
		if not kb.kb.getData( 'sapinfo', 'systeminfo'):
			#import plugins.audit.sapinfo as sapinfo
			# Create a new connection
			conn2 = copy.deepcopy(self._conn)
			try:
				info = conn2.sapinfo()
			except TypeError, e:
				om.out.information('Could not get remote information: ' + str(e))
				return False
			else:
				for key in info.keys():
					info[ key ] = info[ key ].strip()
				kb.kb.save( 'sapinfo', 'systeminfo', info)
				conn2 = None
		
		# KB already with data.
		# All defined keys in Targets must be available
		info = kb.kb.getData('sapinfo','systeminfo')
		for key in targets.keys():
			if key in info.keys():
				# Target can be a list
				for value in targets[key]:
					if not re.compile(value).match(info[key]):
						return False
			else:
					return False
		
		return True
	
# Make CORE World available
sCore = sapytoCore()
